from wpilib.command import Command


def checkIfCanceled(self):
    if self.forceCancel:
        self.forceCancel = False
        return True

    return self._isFinished()


class CancelCommand(Command):
    """
    When this command is run, it cancels the command it was passed, even if that
    command is part of a :class:`wpilib.CommandGroup`.
    """

    def __init__(self, command):
        """
            :param command: The :class:`wpilib.Command` to cancel.
        """
        super().__init__("Cancel %s" % command)

        self.command = command
        if hasattr(self.command, "_isFinished"):
            return

        self.command._isFinished = self.command.isFinished
        self.command.isFinished = checkIfCanceled.__get__(self.command)
        self.command.forceCancel = False

    def initialize(self):
        if self.command.isRunning():
            self.command.forceCancel = True

    def isFinished(self):
        return not self.command.isRunning()



from wpilib.command.commandgroup import CommandGroup


class ConditionalCommand(CommandGroup):
    def __init__(self, name, onTrue=None, onFalse=None):
        super().__init__(name)

        self.onTrue = self._processCommand(onTrue)
        self.onFalse = self._processCommand(onFalse)

    def _processCommand(self, cmd):
        if cmd is None:
            return []

        with self.mutex:
            if self.locked:
                raise ValueError("Cannot add conditions to ConditionalCommand")

            for reqt in cmd.getRequirements():
                self.requires(reqt)

            cmd.setParent(self)
            cmd = CommandGroup.Entry(cmd, CommandGroup.Entry.IN_SEQUENCE, None)
            return [cmd]

    def _initialize(self):
        super()._initialize()

        if self.condition():
            self.commands = self.onTrue
        else:
            self.commands = self.onFalse



import functools
import inspect
import warnings
from typing import Generic, Optional, TypeVar

from networktables import NetworkTables
from ntcore.value import Value

V = TypeVar("V")


class tunable(Generic[V]):
    """
        This allows you to define simple properties that allow you to easily
        communicate with other programs via NetworkTables.
    
        The following example will define a NetworkTable variable at
        ``/components/my_component/foo``::
        
            class MyRobot(magicbot.MagicRobot):
            
                my_component: MyComponent
        
            ...
            
            from magicbot import tunable
        
            class MyComponent:
        
                # define the tunable property
                foo = tunable(True)

                def execute(self):
                
                    # set the variable
                    self.foo = True
                    
                    # get the variable
                    foo = self.foo

        The key of the NetworkTables variable will vary based on what kind of
        object the decorated method belongs to:

        * A component: ``/components/COMPONENTNAME/VARNAME``
        * An autonomous mode: ``/autonomous/MODENAME/VARNAME``
        * Your main robot class: ``/robot/VARNAME``
                    
        .. note:: When executing unit tests on objects that create tunables,
                  you will want to use setup_tunables to set the object up.
                  In normal usage, MagicRobot does this for you, so you don't
                  have to do anything special.
    """

    # the way this works is we use a special class to indicate that it
    # is a tunable, and MagicRobot adds _ntattr and _global_table variables
    # to the class property

    # The tricky bit is that you need to do late binding on these, because
    # the networktables key is not known when the object is created. Instead,
    # the name of the key is related to the name of the variable name in the
    # robot class

    __slots__ = (
        "_ntdefault",
        "_ntsubtable",
        "_ntwritedefault",
        # "__doc__",
        "_mkv",
        "_nt",
    )

    def __init__(
        self,
        default: V,
        *,
        writeDefault: bool = True,
        subtable: Optional[str] = None,
        doc=None
    ) -> None:
        if doc is not None:
            warnings.warn("tunable no longer uses the doc argument", stacklevel=2)

        self._ntdefault = default
        self._ntsubtable = subtable
        self._ntwritedefault = writeDefault
        # self.__doc__ = doc

        self._mkv = Value.getFactory(default)
        self._nt = NetworkTables

    def __get__(self, instance, owner) -> V:
        if instance is not None:
            return instance._tunables[self].value
        return self

    def __set__(self, instance, value) -> None:
        v = instance._tunables[self]
        self._nt._api.setEntryValueById(v._local_id, self._mkv(value))


def setup_tunables(component, cname: str, prefix: Optional[str] = "components") -> None:
    """
        Connects the tunables on an object to NetworkTables.

        :param component:   Component object
        :param cname:       Name of component
        :param prefix:      Prefix to use, or no prefix if None
    
        .. note:: This is not needed in normal use, only useful
                  for testing
    """

    cls = component.__class__

    if prefix is None:
        prefix = "/%s" % cname
    else:
        prefix = "/%s/%s" % (prefix, cname)

    tunables = {}

    for n in dir(cls):
        if n.startswith("_"):
            continue

        prop = getattr(cls, n)
        if not isinstance(prop, tunable):
            continue

        if prop._ntsubtable:
            key = "%s/%s/%s" % (prefix, prop._ntsubtable, n)
        else:
            key = "%s/%s" % (prefix, n)

        ntvalue = NetworkTables.getGlobalAutoUpdateValue(
            key, prop._ntdefault, prop._ntwritedefault
        )
        tunables[prop] = ntvalue

    component._tunables = tunables


def feedback(f=None, *, key: str = None):
    """
    This decorator allows you to create NetworkTables values that are
    automatically updated with the return value of a method.

    ``key`` is an optional parameter, and if it is not supplied,
    the key will default to the method name with a leading ``get_`` removed.
    If the method does not start with ``get_``, the key will be the full
    name of the method.

    The key of the NetworkTables value will vary based on what kind of
    object the decorated method belongs to:

    * A component: ``/components/COMPONENTNAME/VARNAME``
    * Your main robot class: ``/robot/VARNAME``

    The NetworkTables value will be auto-updated in all modes.

    .. warning:: The function should only act as a getter, and must not
                 take any arguments (other than self).

    Example::

        from magicbot import feedback

        class MyComponent:
            navx: ...

            @feedback
            def get_angle(self):
                return self.navx.getYaw()

        class MyRobot(magicbot.MagicRobot):
            my_component: MyComponent

            ...

    In this example, the NetworkTable key is stored at
    ``/components/my_component/angle``.

    .. seealso:: :class:`~wpilib.livewindow.LiveWindow` may suit your needs,
                 especially if you wish to monitor WPILib objects.

    .. versionadded:: 2018.1.0
    """
    if f is None:
        return functools.partial(feedback, key=key)

    if not callable(f):
        raise TypeError(
            "Illegal use of feedback decorator on non-callable {!r}".format(f)
        )
    sig = inspect.signature(f)
    name = f.__name__

    if len(sig.parameters) != 1:
        raise ValueError(
            "{} may not take arguments other than 'self' (must be a simple getter method)".format(
                name
            )
        )

    # Set attributes to be checked during injection
    f._magic_feedback = True
    f._magic_feedback_key = key

    return f


def collect_feedbacks(component, cname: str, prefix="components"):
    """
    Finds all methods decorated with :func:`feedback` on an object
    and returns a list of 2-tuples (method, NetworkTables entry).

    .. note:: This isn't useful for normal use.
    """
    if prefix is None:
        prefix = "/%s" % cname
    else:
        prefix = "/%s/%s" % (prefix, cname)

    nt = NetworkTables.getTable(prefix)
    feedbacks = []

    for name, method in inspect.getmembers(component, inspect.ismethod):
        if getattr(method, "_magic_feedback", False):
            key = method._magic_feedback_key
            if key is None:
                if name.startswith("get_"):
                    key = name[4:]
                else:
                    key = name

            entry = nt.getEntry(key)
            feedbacks.append((method, entry))

    return feedbacks



import contextlib
import inspect
import logging

import hal
import wpilib

from networktables import NetworkTables
from wpilib.shuffleboard import Shuffleboard

from robotpy_ext.autonomous import AutonomousModeSelector
from robotpy_ext.misc import NotifierDelay
from robotpy_ext.misc.orderedclass import OrderedClass
from robotpy_ext.misc.annotations import get_class_annotations

from .magic_tunable import setup_tunables, tunable, collect_feedbacks
from .magic_reset import will_reset_to

__all__ = ["MagicRobot"]


class MagicInjectError(ValueError):
    pass


class MagicRobot(wpilib.SampleRobot, metaclass=OrderedClass):
    """
        Robots that use the MagicBot framework should use this as their
        base robot class. If you use this as your base, you must
        implement the following methods:

        - :meth:`createObjects`
        - :meth:`teleopPeriodic`

        MagicRobot uses the :class:`.AutonomousModeSelector` to allow you
        to define multiple autonomous modes and to select one of them via
        the SmartDashboard/Shuffleboard.

        MagicRobot will set the following NetworkTables variables
        automatically:
        
        - ``/robot/mode``: one of 'disabled', 'auto', 'teleop', or 'test'
        - ``/robot/is_simulation``: True/False
        - ``/robot/is_ds_attached``: True/False
        
    """

    #: Amount of time each loop takes (default is 20ms)
    control_loop_wait_time = 0.020

    #: Error report interval: when an FMS is attached, how often should
    #: uncaught exceptions be reported?
    error_report_interval = 0.5

    #: If True, teleopPeriodic will be called in autonomous mode
    use_teleop_in_autonomous = False

    def robotInit(self):
        """
            .. warning:: Internal API, don't override; use :meth:`createObjects` instead
        """

        self._exclude_from_injection = ["logger", "members"]

        self.__last_error_report = -10

        self._components = []
        self._feedbacks = []
        self._reset_components = []

        # Create the user's objects and stuff here
        self.createObjects()

        # Load autonomous modes
        self._automodes = AutonomousModeSelector("autonomous")

        # Next, create the robot components and wire them together
        self._create_components()

        self.__nt = NetworkTables.getTable("/robot")

        self.__nt_put_is_ds_attached = self.__nt.getEntry("is_ds_attached").setBoolean
        self.__nt_put_mode = self.__nt.getEntry("mode").setString

        self.__nt.putBoolean("is_simulation", self.isSimulation())
        self.__nt_put_is_ds_attached(self.ds.isDSAttached())

        # cache these
        self.__sd_update = wpilib.SmartDashboard.updateValues
        self.__lv_update = wpilib.LiveWindow.updateValues
        self.__sf_update = Shuffleboard.update

        self.watchdog = wpilib.Watchdog(self.control_loop_wait_time, self._loop_overrun)

    def createObjects(self):
        """
            You should override this and initialize all of your wpilib
            objects here (and not in your components, for example). This
            serves two purposes:

            - It puts all of your motor/sensor initialization in the same
              place, so that if you need to change a port/pin number it
              makes it really easy to find it. Additionally, if you want
              to create a simplified robot program to test a specific
              thing, it makes it really easy to copy/paste it elsewhere

            - It allows you to use the magic injection mechanism to share
              variables between components

            .. note:: Do not access your magic components in this function,
                      as their instances have not been created yet. Do not
                      create them either.
        """
        raise NotImplementedError

    def teleopInit(self):
        """
            Initialization code for teleop control code may go here.

            Users may override this method for initialization code which will be
            called each time the robot enters teleop mode.

            .. note:: The ``on_enable`` functions of all components are called
                      before this function is called.
        """
        pass

    def teleopPeriodic(self):
        """
            Periodic code for teleop mode should go here.

            Users should override this method for code which will be called
            periodically at a regular rate while the robot is in teleop mode.

            This code executes before the ``execute`` functions of all
            components are called.

            .. note:: If you want this function to be called in autonomous
                      mode, set ``use_teleop_in_autonomous`` to True in your
                      robot class.
        """
        func = self.teleopPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.warning(
                "Default MagicRobot.teleopPeriodic() method... Override me!"
            )
            func.firstRun = False

    def disabledInit(self):
        """
            Initialization code for disabled mode may go here.

            Users may override this method for initialization code which will be
            called each time the robot enters disabled mode.

            .. note:: The ``on_disable`` functions of all components are called
                      before this function is called.
        """
        pass

    def disabledPeriodic(self):
        """
            Periodic code for disabled mode should go here.

            Users should override this method for code which will be called
            periodically at a regular rate while the robot is in disabled mode.

            This code executes before the ``execute`` functions of all
            components are called.
        """
        func = self.disabledPeriodic.__func__
        if not hasattr(func, "firstRun"):
            self.logger.warning(
                "Default MagicRobot.disabledPeriodic() method... Override me!"
            )
            func.firstRun = False

    def testInit(self):
        """Initialization code for test mode should go here.

        Users should override this method for initialization code which will be
        called each time the robot enters disabled mode.
        """
        pass

    def testPeriodic(self):
        """Periodic code for test mode should go here."""
        pass

    def robotPeriodic(self):
        """
            Periodic code for all modes should go here.

            Users must override this method to utilize it
            but it is not required.

            This function gets called last in each mode.
            You may use it for any code you need to run
            during all modes of the robot (e.g NetworkTables updates)

            The default implementation will update
            SmartDashboard, LiveWindow and Shuffleboard.
        """
        watchdog = self.watchdog
        self.__sd_update()
        watchdog.addEpoch("SmartDashboard")
        self.__lv_update()
        watchdog.addEpoch("LiveWindow")
        self.__sf_update()
        watchdog.addEpoch("Shuffleboard")

    def onException(self, forceReport=False):
        """
            This function must *only* be called when an unexpected exception
            has occurred that would otherwise crash the robot code. Use this
            inside your :meth:`operatorActions` function.

            If the FMS is attached (eg, during a real competition match),
            this function will return without raising an error. However,
            it will try to report one-off errors to the Driver Station so
            that it will be recorded in the Driver Station Log Viewer.
            Repeated errors may not get logged.

            Example usage::

                def teleopPeriodic(self):
                    try:
                        if self.joystick.getTrigger():
                            self.shooter.shoot()
                    except:
                        self.onException()

                    try:
                        if self.joystick.getRawButton(2):
                            self.ball_intake.run()
                    except:
                        self.onException()

                    # and so on...

            :param forceReport: Always report the exception to the DS. Don't
                                set this to True
        """
        # If the FMS is not attached, crash the robot program
        if not self.ds.isFMSAttached():
            raise

        # Otherwise, if the FMS is attached then try to report the error via
        # the driver station console. Maybe.
        now = wpilib.Timer.getFPGATimestamp()

        try:
            if (
                forceReport
                or (now - self.__last_error_report) > self.error_report_interval
            ):
                wpilib.DriverStation.reportError("Unexpected exception", True)
        except:
            pass  # ok, can't do anything here

        self.__last_error_report = now

    @contextlib.contextmanager
    def consumeExceptions(self, forceReport=False):
        """
            This returns a context manager which will consume any uncaught
            exceptions that might otherwise crash the robot.

            Example usage::

                def teleopPeriodic(self):
                    with self.consumeExceptions():
                        if self.joystick.getTrigger():
                            self.shooter.shoot()

                    with self.consumeExceptions():
                        if self.joystick.getRawButton(2):
                            self.ball_intake.run()

                    # and so on...

            :param forceReport: Always report the exception to the DS. Don't
                                set this to True

            .. seealso:: :meth:`onException` for more details
        """
        try:
            yield
        except:
            self.onException(forceReport=forceReport)

    #
    # Internal API
    #

    def autonomous(self):
        """
            MagicRobot will do The Right Thing and automatically load all
            autonomous mode routines defined in the autonomous folder.

            .. warning:: Internal API, don't override
        """

        self.__nt_put_mode("auto")
        self.__nt_put_is_ds_attached(self.ds.isDSAttached())

        self._on_mode_enable_components()

        auto_functions = (
            self._execute_components,
            self._update_feedback,
            self.robotPeriodic,
        )
        if self.use_teleop_in_autonomous:
            auto_functions = (self.teleopPeriodic,) + auto_functions

        self._automodes.run(
            self.control_loop_wait_time,
            auto_functions,
            self.onException,
            watchdog=self.watchdog,
        )

        self._on_mode_disable_components()

    def disabled(self):
        """
            This function is called in disabled mode. You should not
            override this function; rather, you should override the
            :meth:`disabledPeriodic` function instead.

            .. warning:: Internal API, don't override
        """
        watchdog = self.watchdog
        watchdog.reset()

        self.__nt_put_mode("disabled")
        ds_attached = None

        self._on_mode_disable_components()
        try:
            self.disabledInit()
        except:
            self.onException(forceReport=True)
        watchdog.addEpoch("disabledInit()")

        with NotifierDelay(self.control_loop_wait_time) as delay:
            while self.isDisabled():
                if ds_attached != self.ds.isDSAttached():
                    ds_attached = not ds_attached
                    self.__nt_put_is_ds_attached(ds_attached)

                hal.observeUserProgramDisabled()
                try:
                    self.disabledPeriodic()
                except:
                    self.onException()
                watchdog.addEpoch("disabledPeriodic()")

                self._update_feedback()
                self.robotPeriodic()
                watchdog.addEpoch("robotPeriodic()")
                watchdog.disable()

                if watchdog.isExpired():
                    watchdog.printEpochs()

                delay.wait()
                watchdog.reset()

    def operatorControl(self):
        """
            This function is called in teleoperated mode. You should not
            override this function; rather, you should override the
            :meth:`teleopPeriodics` function instead.

            .. warning:: Internal API, don't override
        """
        watchdog = self.watchdog
        watchdog.reset()

        self.__nt_put_mode("teleop")
        # don't need to update this during teleop -- presumably will switch
        # modes when ds is no longer attached
        self.__nt_put_is_ds_attached(self.ds.isDSAttached())

        # initialize things
        self._on_mode_enable_components()

        try:
            self.teleopInit()
        except:
            self.onException(forceReport=True)
        watchdog.addEpoch("teleopInit()")

        observe = hal.observeUserProgramTeleop

        with NotifierDelay(self.control_loop_wait_time) as delay:
            while self.isOperatorControlEnabled():
                observe()
                try:
                    self.teleopPeriodic()
                except:
                    self.onException()
                watchdog.addEpoch("teleopPeriodic()")

                self._execute_components()

                self._update_feedback()
                self.robotPeriodic()
                watchdog.addEpoch("robotPeriodic()")
                watchdog.disable()

                if watchdog.isExpired():
                    watchdog.printEpochs()

                delay.wait()
                watchdog.reset()

        self._on_mode_disable_components()

    def test(self):
        """Called when the robot is in test mode"""
        watchdog = self.watchdog
        watchdog.reset()

        self.__nt_put_mode("test")
        self.__nt_put_is_ds_attached(self.ds.isDSAttached())

        try:
            self.testInit()
        except:
            self.onException(forceReport=True)
        watchdog.addEpoch("testInit()")

        with NotifierDelay(self.control_loop_wait_time) as delay:
            while self.isTest() and self.isEnabled():
                hal.observeUserProgramTest()
                try:
                    self.testPeriodic()
                except:
                    self.onException()
                watchdog.addEpoch("testPeriodic()")

                self._update_feedback()
                self.robotPeriodic()
                watchdog.addEpoch("robotPeriodic()")
                watchdog.disable()

                if watchdog.isExpired():
                    watchdog.printEpochs()

                delay.wait()
                watchdog.reset()

    def _on_mode_enable_components(self):
        # initialize things
        for _, component in self._components:
            on_enable = getattr(component, "on_enable", None)
            if on_enable is not None:
                try:
                    on_enable()
                except:
                    self.onException(forceReport=True)

    def _on_mode_disable_components(self):
        # deinitialize things
        for _, component in self._components:
            on_disable = getattr(component, "on_disable", None)
            if on_disable is not None:
                try:
                    on_disable()
                except:
                    self.onException(forceReport=True)

    def _create_components(self):

        #
        # TODO: Will need to inject into any autonomous mode component
        #       too, as they're a bit different
        #

        # TODO: Will need to order state machine components before
        #       other components just in case

        components = []

        self.logger.info("Creating magic components")

        # Identify all of the types, and create them
        cls = self.__class__

        # - Iterate over class variables with type annotations
        for m, ctyp in get_class_annotations(cls).items():
            # Ignore private variables
            if m.startswith("_"):
                continue

            if hasattr(cls, m):
                attr = getattr(cls, m)
                # If the value given to the variable is an instance of a type and isn't a property
                # raise an error. No double declaring types, e.g foo: type = type
                if isinstance(attr, type) and not isinstance(attr, property):
                    raise ValueError(
                        "%s.%s has two type declarations" % (cls.__name__, m)
                    )
                # Otherwise, skip this set class variable
                continue

            # If the variable has been assigned in __init__ or createObjects, skip it
            if hasattr(self, m):
                continue

            # If the type is not actually a type, give a meaningful error
            if not isinstance(ctyp, type):
                raise TypeError(
                    "%s has a non-type annotation on %s (%r); lone non-injection variable annotations are disallowed, did you want to assign a static variable?"
                    % (cls.__name__, m, ctyp)
                )

            component = self._create_component(m, ctyp)

            # Store for later
            components.append((m, component))

        # - Iterate over set class variables
        for m in self.members:
            if m.startswith("_") or isinstance(getattr(cls, m, None), tunable):
                continue

            ctyp = getattr(self, m)
            if not isinstance(ctyp, type):
                continue

            component = self._create_component(m, ctyp)

            # Store for later
            components.append((m, component))

        # Collect attributes of this robot that are injectable
        self._injectables = {}

        for n in dir(self):
            if (
                n.startswith("_")
                or n in self._exclude_from_injection
                or isinstance(getattr(cls, n, None), tunable)
            ):
                continue

            o = getattr(self, n)

            # Don't inject methods
            # TODO: This could actually be a cool capability..
            if inspect.ismethod(o):
                continue

            self._injectables[n] = o

        # For each new component, perform magic injection
        for cname, component in components:
            setup_tunables(component, cname, "components")
            self._setup_vars(cname, component)
            self._setup_reset_vars(component)

        # Do it for autonomous modes too
        for mode in self._automodes.modes.values():
            mode.logger = logging.getLogger(mode.MODE_NAME)
            setup_tunables(mode, mode.MODE_NAME, "autonomous")
            self._setup_vars(mode.MODE_NAME, mode)

        # And for self too
        setup_tunables(self, "robot", None)
        self._feedbacks += collect_feedbacks(self, "robot", None)

        # Call setup functions for components
        for cname, component in components:
            setup = getattr(component, "setup", None)
            if setup is not None:
                setup()
            # ... and grab all the feedback methods
            self._feedbacks += collect_feedbacks(component, cname, "components")

        # Call setup functions for autonomous modes
        for mode in self._automodes.modes.values():
            if hasattr(mode, "setup"):
                mode.setup()

        self._components = components

    def _create_component(self, name, ctyp):
        # Create instance, set it on self
        component = ctyp()
        setattr(self, name, component)

        # Ensure that mandatory methods are there
        if not callable(getattr(component, "execute", None)):
            raise ValueError(
                "Component %s (%r) must have a method named 'execute'"
                % (name, component)
            )

        # Automatically inject a logger object
        component.logger = logging.getLogger(name)

        self.logger.info("-> %s (class: %s)", name, ctyp.__name__)

        return component

    def _setup_vars(self, cname, component):

        self.logger.debug("Injecting magic variables into %s", cname)

        component_type = type(component)

        # Iterate over variables with type annotations
        for n, inject_type in get_class_annotations(component_type).items():

            # If the variable is private ignore it
            if n.startswith("_"):
                continue

            if hasattr(component_type, n):
                attr = getattr(component_type, n)
                # If the value given to the variable is an instance of a type and isn't a property
                # raise an error. No double declaring types, e.g foo: type = type
                if isinstance(attr, type) and not isinstance(attr, property):
                    raise MagicInjectError(
                        "%s.%s has two type declarations" % (component_type.__name__, n)
                    )
                # Otherwise, skip this set class variable
                continue

            # If the variable has been assigned in __init__, skip it
            if hasattr(component, n):
                continue

            # If the type is not actually a type, give a meaningful error
            if not isinstance(inject_type, type):
                raise TypeError(
                    "Component %s has a non-type annotation on %s (%r); lone non-injection variable annotations are disallowed, did you want to assign a static variable?"
                    % (cname, n, inject_type)
                )

            self._inject(n, inject_type, cname, component)

        # Iterate over static variables
        for n in dir(component):
            # If the variable is private or a proprty, don't inject
            if n.startswith("_") or isinstance(
                getattr(component_type, n, True), property
            ):
                continue

            inject_type = getattr(component, n)

            # If the value assigned isn't a type, don't inject
            if not isinstance(inject_type, type):
                continue

            self._inject(n, inject_type, cname, component)

    def _inject(self, n, inject_type, cname, component):
        # Retrieve injectable object
        injectable = self._injectables.get(n)
        if injectable is None:
            if cname is not None:
                # Try for mangled names
                injectable = self._injectables.get("%s_%s" % (cname, n))

        # Raise error if injectable syntax used but no injectable was found.
        if injectable is None:
            raise MagicInjectError(
                "Component %s has variable %s (type %s), which is not present in %s"
                % (cname, n, inject_type, self)
            )

        # Raise error if injectable declared with type different than the initial type
        if not isinstance(injectable, inject_type):
            raise MagicInjectError(
                "Component %s variable %s in %s are not the same types! (Got %s, expected %s)"
                % (cname, n, self, type(injectable), inject_type)
            )
        # Perform injection
        setattr(component, n, injectable)
        self.logger.debug("-> %s as %s.%s", injectable, cname, n)

    def _setup_reset_vars(self, component):
        reset_dict = {}

        for n in dir(component):
            if isinstance(getattr(type(component), n, True), property):
                continue

            a = getattr(component, n, None)
            if isinstance(a, will_reset_to):
                reset_dict[n] = a.default

        if reset_dict:
            component.__dict__.update(reset_dict)
            self._reset_components.append((reset_dict, component))

    def _update_feedback(self):
        for method, entry in self._feedbacks:
            try:
                value = method()
            except:
                self.onException()
                continue
            entry.setValue(value)
        self.watchdog.addEpoch("@magicbot.feedback")

    def _execute_components(self):
        for name, component in self._components:
            try:
                component.execute()
            except:
                self.onException()
            self.watchdog.addEpoch(name)

        for reset_dict, component in self._reset_components:
            component.__dict__.update(reset_dict)

    def _loop_overrun(self):
        # TODO: print something here without making it extremely annoying
        pass



from .magicrobot import MagicRobot
from .magic_tunable import feedback, tunable
from .magic_reset import will_reset_to

from .state_machine import (
    AutonomousStateMachine,
    StateMachine,
    default_state,
    state,
    timed_state,
)



import importlib
import inspect
import logging
import os
from glob import glob

import hal
import wpilib

from ..misc.precise_delay import NotifierDelay

logger = logging.getLogger("autonomous")


class AutonomousModeSelector:
    """
        This object loads all modules in a specified python package, and tries
        to automatically discover autonomous modes from them. Each module is 
        added to a ``SendableChooser`` object, which allows the user to select
        one of them via the SmartDashboard/SFX.
        
        Autonomous mode objects must implement the following functions:
        
        - ``on_enable`` - Called when autonomous mode is initially enabled
        - ``on_disable`` - Called when autonomous mode is no longer active
        - ``on_iteration`` - Called for each iteration of the autonomous control loop
        
        Your autonomous object may have the following attributes:
        
        - ``MODE_NAME`` - The name of the autonomous mode to display to users
        - ``DISABLED`` - If True, don't allow this mode to be selected
        - ``DEFAULT`` - If True, this is the default autonomous mode selected
        
        AutonomousModeSelector can be easily used from either ``IterativeRobot``
        or from ``SampleRobot``. For ``IterativeRobot``, the simplest usage would
        look like so::
        
            class MyRobot(wpilib.IterativeRobot):
            
                def robotInit(self):
                    self.automodes = AutonomousModeSelector('autonomous')
            
                def autonomousPeriodic(self):
                    self.automodes.run()
                    
        
        For ``SampleRobot``, for the simplest usage you would do this::
        
            class MyRobot(wpilib.SampleRobot):
            
                def robotInit(self):
                    self.automodes = AutonomousModeSelector('autonomous')
            
                def autonomous(self):
                    self.automodes.run()
       
        If you use AutonomousModeSelector, you may also be interested in
        the autonomous state machine helper (:class:`.StatefulAutonomous`).
        
        Check out the samples in our github repository that show some basic
        usage of ``AutonomousModeSelector``.
        
        .. note:: If you use AutonomousModeSelector, then you should add
                  ``robotpy_ext.autonomous.selector_tests`` to your pyfrc
                  unit tests like so::
                  
                      from robotpy_ext.autonomous.selector_tests import *
    """

    def __init__(self, autonomous_pkgname, *args, **kwargs):
        """
            :param autonomous_pkgname: Module to load autonomous modes from
            :param args: Args to pass to created autonomous modes
            :param kwargs: Keyword args to pass to created autonomous modes
        """

        self.ds = wpilib.DriverStation.getInstance()
        self.modes = {}
        self.active_mode = None

        logger.info("Begin initializing autonomous mode switcher")

        # load all modules in specified module
        modules = []

        try:
            autonomous_pkg = importlib.import_module(autonomous_pkgname)
        except ImportError as e:
            if e.name not in [autonomous_pkgname, autonomous_pkgname.split(".")[0]]:
                raise

            # Don't kill the robot because they didn't create an autonomous package
            logger.warning("Cannot load the '%s' package", autonomous_pkgname)
        else:
            if hasattr(autonomous_pkg, "__file__"):
                modules_path = os.path.dirname(os.path.abspath(autonomous_pkg.__file__))
                modules = glob(os.path.join(modules_path, "*.py"))

        for module_filename in modules:

            module = None
            module_name = os.path.basename(module_filename[:-3])

            if module_name in ["__init__"]:
                continue

            try:
                module = importlib.import_module("." + module_name, autonomous_pkgname)
                # module = imp.load_source('.' + module_name, module_filename)
            except:
                if not self.ds.isFMSAttached():
                    raise

            #
            # Find autonomous mode classes in the modules that are present
            # -> note that we actually create the instance of the objects here,
            #    so that way we find out about any errors *before* we get out
            #    on the field..

            for name, obj in inspect.getmembers(module, inspect.isclass):
                mode_name = getattr(obj, "MODE_NAME", None)
                if mode_name is not None:

                    # don't allow the driver to select this mode
                    if getattr(obj, "DISABLED", False):
                        logger.warning(
                            "autonomous mode %s is marked as disabled", obj.MODE_NAME
                        )
                        continue

                    try:
                        instance = obj(*args, **kwargs)
                    except:

                        if not self.ds.isFMSAttached():
                            raise
                        else:
                            continue

                    if mode_name in self.modes:
                        if not self.ds.isFMSAttached():
                            raise RuntimeError(
                                "Duplicate name %s in %s" % (mode_name, module_filename)
                            )

                        logger.error(
                            "Duplicate name %s specified by object type %s in module %s",
                            mode_name,
                            name,
                            module_filename,
                        )
                        self.modes[name + "_" + module_filename] = instance
                    else:
                        self.modes[instance.MODE_NAME] = instance

        # now that we have a bunch of valid autonomous mode objects, let
        # the user select one using the SmartDashboard.

        # SmartDashboard interface
        self.chooser = wpilib.SendableChooser()

        default_modes = []
        mode_names = []

        logger.info("Loaded autonomous modes:")
        for k, v in sorted(self.modes.items()):

            if getattr(v, "DEFAULT", False):
                logger.info(" -> %s [Default]", k)
                self.chooser.setDefaultOption(k, v)
                default_modes.append(k)
            else:
                logger.info(" -> %s", k)
                self.chooser.addOption(k, v)

            mode_names.append(k)

        if len(self.modes) == 0:
            logger.warning("-- no autonomous modes were loaded!")

        self.chooser.addOption("None", None)

        if len(default_modes) == 0:
            self.chooser.setDefaultOption("None", None)
        elif len(default_modes) != 1:
            if not self.ds.isFMSAttached():
                raise RuntimeError(
                    "More than one autonomous mode was specified as default! (modes: %s)"
                    % (", ".join(default_modes))
                )

        # must PutData after setting up objects
        wpilib.SmartDashboard.putData("Autonomous Mode", self.chooser)

        # XXX: Compatibility with the FRC dashboard
        wpilib.SmartDashboard.putStringArray("Auto List", mode_names)

        logger.info("Autonomous switcher initialized")

    def run(
        self,
        control_loop_wait_time=0.020,
        iter_fn=None,
        on_exception=None,
        watchdog: wpilib.Watchdog = None,
    ):
        """
            This function does everything required to implement autonomous
            mode behavior. You should call this from your autonomous mode
            function -- ``autonomousPeriodic`` in :class:`.IterativeRobot`,
            or ``autonomous`` in :class:`.SampleRobot`.
            
            This function will NOT exit until autonomous mode has ended. If
            you need to execute code in all autonomous modes, pass a function
            or list of functions as the ``iter_fn`` parameter, and they will be
            called once per autonomous mode iteration.
                          
            :param control_loop_wait_time: Amount of time between iterations
            :param iter_fn: Called at the end of every iteration while
                            autonomous mode is executing
            :param on_exception: Called when an uncaught exception is raised,
                                 must take a single keyword arg "forceReport"
            :param watchdog: a WPILib Watchdog to feed every iteration
        """
        if watchdog:
            watchdog.reset()

        logger.info("Begin autonomous")

        if iter_fn is None:
            iter_fn = lambda: None
        if not isinstance(iter_fn, (list, tuple)):
            iter_fn = (iter_fn,)

        if on_exception is None:
            on_exception = self._on_exception

        # keep track of how much time has passed in autonomous mode
        timer = wpilib.Timer()
        timer.start()

        try:
            self._on_autonomous_enable()
        except:
            on_exception(forceReport=True)
        if watchdog:
            watchdog.addEpoch("auto on_enable")

        #
        # Autonomous control loop
        #

        observe = hal.observeUserProgramAutonomous

        with NotifierDelay(control_loop_wait_time) as delay:
            while self.ds.isAutonomousEnabled():
                observe()
                try:
                    self._on_iteration(timer.get())
                except:
                    on_exception()
                if watchdog:
                    watchdog.addEpoch("auto on_iteration")

                for fn in iter_fn:
                    fn()

                if watchdog:
                    watchdog.addEpoch("robotPeriodic()")
                    watchdog.disable()

                    if watchdog.isExpired():
                        watchdog.printEpochs()

                delay.wait()
                if watchdog:
                    watchdog.reset()

        #
        # Done with autonomous, finish up
        #

        try:
            self._on_autonomous_disable()
        except:
            on_exception(forceReport=True)

        logger.info("Autonomous mode ended")

    #
    #   Internal methods used to implement autonomous mode switching, and
    #   are called automatically
    #

    def _on_autonomous_enable(self):
        """Selects the active autonomous mode and enables it"""

        # XXX: FRC Dashboard compatibility
        # -> if you set it here, you're stuck using it. The FRC Dashboard
        #    doesn't seem to have a default (nor will it show a default),
        #    so the key will only get set if you set it.
        auto_mode = wpilib.SmartDashboard.getString("Auto Selector", None)
        if auto_mode is not None and auto_mode in self.modes:
            logger.info("Using autonomous mode set by LabVIEW dashboard")
            self.active_mode = self.modes[auto_mode]
        else:
            self.active_mode = self.chooser.getSelected()

        if self.active_mode is not None:
            logger.info("Enabling '%s'", self.active_mode.MODE_NAME)
            self.active_mode.on_enable()
        else:
            logger.warning(
                "No autonomous modes were selected, not enabling autonomous mode"
            )

    def _on_autonomous_disable(self):
        """Disable the active autonomous mode"""
        if self.active_mode is not None:
            logger.info("Disabling '%s'" % self.active_mode.MODE_NAME)
            self.active_mode.on_disable()

        self.active_mode = None

    def _on_iteration(self, time_elapsed):
        """Run the code for the current autonomous mode"""
        if self.active_mode is not None:
            self.active_mode.on_iteration(time_elapsed)

    def _on_exception(self, forceReport=False):
        if not self.ds.isFMSAttached():
            raise



import pyfrc.config
import pytest
from networktables.util import ChooserControl

autonomous_seconds = 15
_gsms = pyfrc.config.config_obj["pyfrc"]["game_specific_messages"]


@pytest.mark.parametrize("gamedata", _gsms or [""])
def test_all_autonomous(control, fake_time, robot, gamedata):
    """
        This test runs all possible autonomous modes that can be selected
        by the autonomous switcher.
        
        This should work for most robots. If it doesn't work for yours, 
        and it's not a code issue with your robot, please file a bug on
        github.
    """

    class AutonomousTester:
        def __init__(self):
            self.initialized = False
            self.init_time = None
            self.chooser = None

            self.state = "auto"
            self.currentChoice = None
            self.until = None

        def initialize_chooser(self, tm):

            if self.chooser is None:
                self.chooser = ChooserControl("Autonomous Mode")

            self.choices = self.chooser.getChoices()
            if len(self.choices) == 0:
                return False

            self.state = "disabled"
            self.currentChoice = -1
            self.until = tm
            self.init_time = tm
            self.initialized = True
            return True

        def on_step(self, tm):

            if not self.initialized:
                if not self.initialize_chooser(tm):
                    assert (
                        tm < 10
                    ), "Robot didn't create a chooser within 10 seconds, probably an error"
                    return True

            if self.state == "auto":
                if tm >= self.until:
                    self.until = tm + 1
                    self.state = "disabled"
                    control.set_operator_control(enabled=False)

            elif self.state == "disabled":
                if tm >= self.until:

                    control.set_autonomous()

                    self.state = "auto"
                    self.until = tm + autonomous_seconds
                    self.currentChoice += 1
                    if self.currentChoice >= len(self.choices):
                        return False

                    self.chooser.setSelected(self.choices[self.currentChoice])

            return True

    control.game_specific_message = gamedata
    controller = control.run_test(AutonomousTester)

    # Make sure they ran for the correct amount of time
    assert int(fake_time.get()) == int(
        len(controller.choices) * (autonomous_seconds + 1) + controller.init_time
    )



"""
These functions can be used to make programming CommandGroups much more
intuitive. For more information, check each method's docstring.
"""

from wpilib.command import Command, CommandGroup
from commandbased.conditionalcommand import ConditionalCommand
from commandbased.cancelcommand import CancelCommand

import inspect

__all__ = ["IF", "ELIF", "ELSE", "WHILE", "RETURN", "BREAK"]


def _getCommandGroup():
    """
    Does some rather ugly stack inspection to find out which CommandGroup the
    calling function was triggered from.
    """
    # https://stackoverflow.com/a/14694234
    stack = inspect.stack()
    frame = stack[2].frame
    while "self" not in frame.f_locals:
        frame = frame.f_back
        if frame is None:
            raise ValueError("Could not find calling class for %s" % stack[1].function)

    cg = frame.f_locals["self"]
    if not isinstance(cg, CommandGroup):
        raise ValueError(
            "%s may not be used outside of a CommandGroup" % stack[1].function
        )

    return cg


def _getSource(parent):
    """Finds the highest level CommandGroup of parent"""

    try:
        return parent._source
    except AttributeError:
        return parent


def _buildCommandGroup(func, parent):
    """Turns the given function into a full CommandGroup."""

    source = _getSource(parent)

    cg = CommandGroup(func.__name__)
    cg._source = source
    func(cg)

    return cg


def _restartWhile(self):
    """
    Replaces isFinished for a ConditionalCommand in a WHILE loop, to keep it
    running.
    """
    if self.forceCancel:
        self.forceCancel = False
        return True

    finished = super(ConditionalCommand, self).isFinished()

    if finished:
        if self.condition():
            self.currentCommandIndex = None
            return False

    return finished


def _popIfStack(cg):
    """
    We buffer conditionals until the last moment so we don't have trouble with
    Commands being locked when they're added to a CommandGroup.
    """
    if cg._ifStack:
        top = cg._ifStack.pop(0)
        cmd = None
        for x in reversed(cg._ifStack):
            if x[0]:
                cmd = ConditionalCommand("flowcontrolELIF", x[1], cmd)
                cmd.condition = x[0]

            else:
                cmd = x[1]

        cmd = ConditionalCommand("flowcontrolIF", top[1], cmd)
        cmd.condition = top[0]

        cg._addSequential(cmd)
        cg._ifStack = None


# These _hook methods ensure we always add our buffered conditions
def _hookSequential(self, cmd, timeout=None):
    _popIfStack(self)
    self._addSequential(cmd, timeout)


def _hookParallel(self, cmd, timeout=None):
    _popIfStack(self)
    self._addParallel(cmd, timeout)


def _hookStart(self):
    _popIfStack(self)
    self._start()


def _hookParent(self, parent):
    _popIfStack(self)
    self._setParent(parent)


def _hookCommandGroup(cg):
    """Override some methods of the CommandGroup to add buffered commands."""

    try:
        cg._ifStack
        return
    except AttributeError:
        pass

    cg._addSequential = cg.addSequential
    cg._addParallel = cg.addParallel
    cg._start = cg.start
    cg._setParent = cg.setParent

    cg.addSequential = _hookSequential.__get__(cg)
    cg.addParallel = _hookParallel.__get__(cg)
    cg.start = _hookStart.__get__(cg)
    cg.setParent = _hookParent.__get__(cg)


def IF(condition):
    """
    Use as a decorator for a function. That function will be placed into a
    CommandGroup and run inside a ConditionalCommand with the given condition.
    The decorated function must accept one positional argument that will be used
    as its 'self'.
    """

    def flowcontrolIF(func):
        parent = _getCommandGroup()
        _hookCommandGroup(parent)

        cg = _buildCommandGroup(func, parent)

        try:
            _popIfStack(parent)
        except AttributeError:
            pass

        parent._ifStack = [(condition, cg)]

    return flowcontrolIF


def ELIF(condition):
    """
    Use as a decorator for a function. That function will be placed into a
    CommandGroup which will be triggered by a ConditionalCommand that uses the
    passed condition. That ConditionalCommand will then be added as the onFalse
    for the ConditionalCommand created by a previous IF or ELIF.
    """

    def flowcontrolELIF(func):
        parent = _getCommandGroup()

        cg = _buildCommandGroup(func, parent)

        try:
            parent._ifStack.append((condition, cg))
        except AttributeError:
            raise ValueError("Cannot use ELIF without IF")

    return flowcontrolELIF


def ELSE(func):
    """
    Use as a decorator for a function. That function will be placed into a
    CommandGroup which will be added as the onFalse for the ConditionalCommand
    created by a previous IF or ELIF.
    """

    parent = _getCommandGroup()
    cg = _buildCommandGroup(func, parent)

    try:
        parent._ifStack.append((None, cg))
    except AttributeError:
        raise ValueError("Cannot use ELSE without IF")

    _popIfStack(parent)


def WHILE(condition):
    """
    Use as a decorator for a function. That function will be placed into a
    CommandGroup, which will be added to a ConditionalCommand. It will be
    modified to restart itself automatically.
    """

    def flowcontrolWHILE(func):
        parent = _getCommandGroup()
        source = _getSource(parent)

        try:
            parentLoop = source._currentLoop
        except AttributeError:
            parentLoop = None

        cg = CommandGroup(func.__name__)
        cg._source = source

        # Set the current loop for any BREAK statements
        source._currentLoop = cg
        func(cg)
        source._currentLoop = parentLoop

        def cancelLoop(self):
            self.getGroup().forceCancel = True

        end = Command("END WHILE")
        end.initialize = cancelLoop.__get__(end)

        cond = ConditionalCommand("flowcontrolWHILE", cg, end)
        cond.condition = condition
        cond.forceCancel = False
        cond.isFinished = _restartWhile.__get__(cond)
        cond._parentLoop = parentLoop

        parent.addSequential(cond)

        cg.conditionalCommand = cond

    return flowcontrolWHILE


def RETURN():
    """
    Calling this function will end the source CommandGroup immediately.
    """

    parent = _getCommandGroup()
    source = _getSource(parent)

    parent.addSequential(CancelCommand(source))


def BREAK(steps=1):
    """
    Calling this function will end the loop that contains it. Pass an integer to
    break out of that number of nested loops.
    """

    if steps < 1:
        raise ValueError("Steps to BREAK cannot be < 1")

    parent = _getCommandGroup()
    source = _getSource(parent)

    try:
        loop = source._currentLoop
    except AttributeError:
        raise ValueError("Cannot BREAK outside of a loop")

    if loop is None:
        raise ValueError("Cannot BREAK outside of a loop")

    # We can't use CancelCommand here, because the conditionalCommand attribute
    # isn't yet bound to the CommandGroup, so we find it at initialization time
    # instead
    def cancelLoop():
        nonlocal loop, steps
        step = 1
        while steps > step:
            loop = loop.conditionalCommand._parentLoop

            if loop is None:
                raise ValueError(
                    "BREAK %i not possible with loop depth %i" % (steps, step)
                )

            step += 1

        loop.conditionalCommand.forceCancel = True

    breakLoop = Command("flowcontrolBREAK")
    breakLoop.initialize = cancelLoop
    parent.addSequential(breakLoop)
